--CORE OF ADDON
-- =========================
local testData, core = ...;
TAB = LibStub("AceAddon-3.0"):NewAddon(core, "Talent Action Bars", "AceConsole-3.0", "AceComm-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0", "AceSerializer-3.0")
local TAB	= LibStub("AceAddon-3.0"):GetAddon("Talent Action Bars")
local AceGUI = LibStub("AceGUI-3.0")
local LibDeflate = LibStub:GetLibrary("LibDeflate")
local setupPick, enabledCHB, setupName, editbox;


--COMMAND OPTIONS
-- =========================
local options = {
	name = 'Talent Action Bars',
	type = 'group',
	args = {
		enable = {
			type = 'execute',
			name = "Enable/Disable",
			desc = "Activate main function of addon",
			func = function() TAB:ToggleMiniMapIcon() end
		},
		version = {
			type = 'execute',
			name = "Version",
			desc = "Show version of addon",
			func = function() TAB:Print("v" .. TAB.db.global.version) end
		},
	}
}


--ADDON OPTIONS IN INTERFACE
-- =========================
local settings = {
	name = 'Talent Action Bars Settings',
	type = 'group',
	args = {
		minimap = {
			order = 0,
			type = "toggle",
			name = "Hide Minimap icon",
			desc = "Hide Minimap icon",
			get = function() return TAB.db.char.minimap.hide; end,
			set = function(info, v) TAB:ToggleMinimap(); end,
		},
		enabled = {
			order = 1,
			type = "toggle",
			name = "Enabled",
			desc = "Activate main function of addon",
			get = function() return TAB.db.char.enabled; end,
			set = function(info, v) TAB:ToggleMiniMapIcon(); end,
		},
		ignoreMacro = {
			order = 2,
			type = "toggle",
			name = "Ignore macro change",
			desc = "Always set macro from position where was last seen if position still exists.",
			get = function() return TAB.db.global.ignoreMacro; end,
			set = function(info, v) TAB.db.global.ignoreMacro = v end,
		},
		changeBinds = {
			order = 3,
			type = "toggle",
			name = "Change Binds",
			desc = "Allow to change also binds on bars.",
			get = function() return TAB.db.global.binds; end,
			set = function(info, v) TAB.db.global.binds = v end,
		},
		barsHeader = {
			order = 4,
			type = "header",
			name = "Bars",
		},
		bars = {
			order = 5,
			type = "multiselect",
			name = "Bars",
			desc = "Select bars on which you want apply changes.",
			values = function() return {"Bar 1", "Bar 2", "Bar 3", "Bar 4", "Bar 5", "Bar 6", "Bar 7", "Bar 8", "Bar 9", "Bar 10"}; end,
			get = function(info, index) return TAB.db.char.actionBars[index]; end,
			set = function(info, index, value) TAB.db.char.actionBars[index] = value; end,
		},
		setupsHeader = {
			order = 6,
			type = "header",
			name = "Saved Setups",
		},
		setups = {
			order = 7,
			type = "execute",
			name = "Setups Menu",
			desc = "Menu for saved setups.",
			func = function() TAB:CreateSetupMenu(); end,
		},
		debugHeader = {
			order = 8,
			type = "header",
			name = "Debug",
		},
		debug = {
			order = 9,
			type = "toggle",
			name = "Debug",
			desc = "Allow debug texts",
			get = function() return TAB.db.global.debug; end,
			set = function(info, v) TAB.db.global.debug = v; end,
		},
	}
}


--SCHEME FOR DB
-- =========================
local defaults = {
	global = {
		version = 0,
		ignoreMacro = false,
		debug = false
	},
	char = {
		enabled = true,
		actionBars = {true, true, true, true, true, true, true, true, true, true},
		talentStrings = {},
		bars = {},
		barNames = {},
		barsBackup = {},
		binds = false,
		minimap = {hide = false}
	},
}


--Init function
-- =========================
function TAB:OnInitialize()
	TAB.db = LibStub("AceDB-3.0"):New("TalentActionBarsDB", defaults);
	LibStub("AceConfig-3.0"):RegisterOptionsTable("TalentActionBars", options, {"talentActionBars", "tab"});
	LibStub("AceConfig-3.0"):RegisterOptionsTable("TalentActionBars-Settings", settings);
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("TalentActionBars-Settings", "Talent Action Bars");
	self:RegisterEvent("CHARACTER_POINTS_CHANGED", function(...) self:TalentsChangeCallback(...) end);
	self:RegisterEvent("CONFIRM_TALENT_WIPE", function(...) self:TalentsRemovedCallback(...) end);
	self:RegisterComm("TABVersionCheck", function(...) self:OnCommReceived(...) end);
	self:ConstructMiniMapIcon();
	self.versionWarrning = false;
	self:FillNames();
	self:CheckVersion();
end


--MiniMap Icon Definition
-- =========================
function TAB:ConstructMiniMapIcon() 
	LibStub("LibDBIcon-1.0"):Register("TalentActionBars", LibStub("LibDataBroker-1.1"):NewDataObject("TalentActionBars",
	{
		type = "data source",
		text = addonName,
		icon = "Interface\\Icons\\inv_misc_punchcards_blue",
		OnClick = function(...) TAB:ToggleMiniMapIcon(); end,
		OnTooltipShow = function(tooltip)
			tooltip:AddLine("Talent Action Bars",1,1,1);
			tooltip:Show();
		end
	}), self.db.char.minimap);
	if LibDataBroker then
		LibDataBroker:NewDataObject("Talent Action Bars", {
			type = "launcher",
			icon = "Interface\\Icons\\inv_misc_punchcards_blue",
			OnClick = function(...) TAB:ToggleMiniMapIcon(); end,
		})
	end
	
	if not self.db.char.minimap.hide then
		if TAB.db.char.enabled then
			LibStub("LibDBIcon-1.0"):GetMinimapButton("TalentActionBars").icon:SetDesaturated();
		else
			LibStub("LibDBIcon-1.0"):GetMinimapButton("TalentActionBars").icon:SetDesaturated(1);
		end
	end
end


--Enable/Disable addon function
-- =========================
function TAB:ToggleMiniMapIcon()
	TAB.db.char.enabled = not TAB.db.char.enabled; 
	if TAB.db.char.enabled then
		LibStub("LibDBIcon-1.0"):GetMinimapButton("TalentActionBars").icon:SetDesaturated();
		self:RegisterEvent("CHARACTER_POINTS_CHANGED", function(...) self:TalentsChangeCallback(...) end);
		self:RegisterEvent("CONFIRM_TALENT_WIPE", function(...) self:TalentsRemovedCallback(...) end);
		TAB:Print("Addon enabled!");
	else
		LibStub("LibDBIcon-1.0"):GetMinimapButton("TalentActionBars").icon:SetDesaturated(1);
		self:UnregisterEvent("CHARACTER_POINTS_CHANGED");
		self:UnregisterEvent("CONFIRM_TALENT_WIPE");
		TAB:Print("|cffff0000Addon disabled!");
	end
end


--Enable/Disable minimap icon
-- =========================
function TAB:ToggleMinimap()
	local icon = LibStub("LibDBIcon-1.0")
	if(self.db.char.minimap.hide) then
		self.db.char.minimap.hide = false;
		icon:Show("TalentActionBars");
	else
		self.db.char.minimap.hide = true;
		icon:Hide("TalentActionBars");
	end
end


--GUI menu for saved setups
-- =========================
function TAB:CreateSetupMenu()
	self.ui = {};
	self.ui.setupMenu = AceGUI:Create("Frame");
	self.ui.setupMenu:SetTitle("Setups Menu");
	self.ui.setupMenu:SetStatusText("");
	self.ui.setupMenu:SetLayout("Flow");
	self.ui.setupMenu:SetWidth(365);
	self.ui.setupMenu:SetHeight(200);
	self.ui.setupMenu:EnableResize(false);
	self.ui.setupMenu.frame:EnableMouse(true);
	self.ui.setupMenu.frame:SetUserPlaced(true);
	
	setupPick = AceGUI:Create("Dropdown");
	setupPick:SetLabel("Setup");
	setupPick:SetWidth(320);
	setupPick:SetCallback("OnValueChanged", function(obj, func, value) TAB:LoadSetup(); end);
	self.ui.setupMenu:AddChild(setupPick);
	
	enabledCHB = AceGUI:Create("CheckBox");
	enabledCHB:SetLabel("Check only");
	enabledCHB:SetWidth(200);
	enabledCHB:SetValue(false);
	--enabledCHB:SetCallback("OnValueChanged", function(obj, func, value) --[[TODO--]] end); 
	self.ui.setupMenu:AddChild(enabledCHB);
	
	setupName = AceGUI:Create("EditBox");
	setupName:SetLabel("Setup Name");
	setupName:SetWidth(285);
	setupName:SetCallback("OnEnterPressed", function(widget, event, text) TAB:SaveSetupName(text); end);
	self.ui.setupMenu:AddChild(setupName);
	
	local checkSetup = AceGUI:Create("Icon");
	checkSetup:SetImage("Interface\\Icons\\inv_letter_13");
	checkSetup:SetWidth(20);
	checkSetup:SetHeight(20);
	checkSetup:SetImageSize(15,15);
	checkSetup:SetCallback("OnClick", function() TAB:CopyURL(); end);
	self.ui.setupMenu:AddChild(checkSetup);
	
	local removeSetup = AceGUI:Create("Icon");
	removeSetup:SetImage("Interface\\Icons\\spell_chargenegative");
	removeSetup:SetWidth(20);
	removeSetup:SetHeight(20);
	removeSetup:SetImageSize(15,15);
	removeSetup:SetCallback("OnClick", function() --[[TODO--]] end);
	removeSetup:SetDisabled(true);
	self.ui.setupMenu:AddChild(removeSetup);
	
	TAB:LoadSetups();
	_G["TABSetupMenu"] = self.ui.setupMenu.frame;
    tinsert(UISpecialFrames, "TABSetupMenu");
end


--GUI for copy
-- =========================
function TAB:CreateCopyFrame(text)
	self.ui = {};
	self.ui.copyFrame = AceGUI:Create("Frame");
	self.ui.copyFrame:SetTitle("Copy");
	self.ui.copyFrame:SetStatusText("");
	self.ui.copyFrame:SetLayout("Flow");
	self.ui.copyFrame:SetWidth(400);
	self.ui.copyFrame:SetHeight(150);
	self.ui.copyFrame:EnableResize(false);
	self.ui.copyFrame.frame:EnableMouse(true);
	self.ui.copyFrame.frame:SetUserPlaced(true);
	
	_G["TABCopyFrame"] = self.ui.copyFrame.frame;
    tinsert(UISpecialFrames, "TABCopyFrame");
	
	editbox = AceGUI:Create("MultiLineEditBox");
	editbox:SetLabel("");
	editbox:DisableButton(true);
	editbox:SetWidth(380);
	editbox:SetHeight(70);
	self.ui.copyFrame:AddChild(editbox);
	
	editbox:SetText(text);
	editbox:HighlightText(0, #text);
	editbox:SetFocus();
end


--Load saved setups
-- =========================
function TAB:LoadSetups()
	setupPick:SetValue(nil);
	setupPick:SetList({});
	for i = 1, #TAB.db.char.barNames do
		setupPick:AddItem(i, TAB.db.char.barNames[i]);
	end
	local index = self:IndexOfTalentString(self:GetTalentString());
	setupPick:SetValue(index);
	setupName:SetText(TAB.db.char.barNames[index]);
end


--Change name of setup
-- =========================
function TAB:SaveSetupName(text)
	local index = setupPick:GetValue();
	if(TAB.db.char.barNames[index] ~= nil) then
		TAB.db.char.barNames[index] = text;
		TAB:LoadSetups();
	end
end


--Pick different setup
-- =========================
function TAB:LoadSetup()
	local index = setupPick:GetValue();
	setupName:SetText(TAB.db.char.barNames[index]);
	if(enabledCHB.GetValue()) then
		TAB:LoadBars(index);
	end
end


--Copy Wowhead URL of actual setup
-- =========================
function TAB:CopyURL()
	local talentStringRaw = TAB:GetTalentString();
	local class = UnitClass("player");
	talentStringRaw = string.gsub(talentStringRaw, "-1", "");
	talentStringRaw = string.gsub(talentStringRaw, "-2", "-");
	talentStringRaw = string.gsub(talentStringRaw, "-3", "-");
	TAB:CreateCopyFrame("https://tbc.wowhead.com/talent-calc/" .. class:lower() .. "/" .. talentStringRaw);
end


--Callback function for talent picked
-- =========================
function TAB:TalentsChangeCallback(...)
	if(UnitCharacterPoints("player") == 0) then
		TAB:LoadBars();
	end
end


--Callback function for talents reset
-- =========================
function TAB:TalentsRemovedCallback(...)
	self:SaveBars();
end

--Actual Talent String
-- =========================
function TAB:GetTalentString()
	local tempString = "";
	for i = 1, 3 do
		tempString =  tempString .. "-" .. tostring(i);
		for j = 1, GetNumTalents(i) do
			name, iconPath, tier, column, currentRank, maxRank, isExceptional, meetsPrereq = GetTalentInfo(i, j);
			tempString =  tempString .. tostring(currentRank);
		end
	end
	return tempString;
end


--Save Current Bars
-- =========================
function TAB:SaveBars()
	local talents = self:GetTalentString();
	local talentIndex = self:FindIndex(TAB.db.char.talentStrings, talents);
	if(talentIndex == nil) then
		table.insert(TAB.db.char.talentStrings, talents);
		talentIndex = #TAB.db.char.talentStrings;
		table.insert(TAB.db.char.barNames, "Bars #" .. talentIndex);
	end
	
	TAB.db.char.barsBackup[talentIndex] = TAB.db.char.bars[talentIndex];
	TAB.db.char.bars[talentIndex] = {};
	
	for i = 1, 120 do
		local type, id, subType = GetActionInfo(i)
		local node = {
			type = nil,
			id = nil,
			name = "",
			index = i,
			bind = nil
		};
		if(type and id) then
			if( type == "companion" ) then
				node.type = type;
				node.id = id;
				node.name = name;
				node.subType = subType;
			elseif( type == "item" ) then
				node.type = type;
				node.id = id;
				node.name = (GetItemInfo(id));
			elseif( type == "spell" and id > 0 ) then
				local spell = GetSpellInfo(id)
				if( spell ) then
					node.type = type;
					node.id = id;
					node.name = spell;
					node.rank = GetSpellSubtext(id) or "";
				end
			elseif( type == "macro" ) then
				local name, icon, macro = GetMacroInfo(id)
				if( name and icon and macro ) then
					node.type = type;
					node.id = id;
					node.name = name;
					node.icon = icon;
					node.macro = macro;
				end
			end
		end
		if(TAB.db.char.bars ~= nil and TAB.db.char.bars[talentIndex] ~= nil) then
			table.insert(TAB.db.char.bars[talentIndex],node);
		end
	end
	
	self:Print("Action Bars are saved.")
end


--Load Current Bars
-- =========================
function TAB:LoadBars(talentIndex)
	if(talentIndex == nil) then
		talentIndex = self:IndexOfTalentString(self:GetTalentString());
	end
	
	if(TAB.db.char.bars[talentIndex] == nil) then
		self:Print("Talent tree was not saved yet.")
	else
		for i = 1, #TAB.db.char.bars[talentIndex] do
			local bar = math.ceil(TAB.db.char.bars[talentIndex][i].index / 12);
			if(TAB.db.char.actionBars[bar]) then
				local type, id, subType = GetActionInfo(TAB.db.char.bars[talentIndex][i].index);
				if(not(TAB.db.char.bars[talentIndex][i].type == type and TAB.db.char.bars[talentIndex][i].id == id)) then
					TAB:ChangeBar(TAB.db.char.bars[talentIndex][i]);
					TAB:Debug("Action bar no. " .. TAB.db.char.bars[talentIndex][i].index .. " was changed.");
				end
			end
		end
		
		self:Print("Action Bars are loaded.")
	end
end


--Change Bar
-- =========================
function TAB:ChangeBar(bar)
	ClearCursor();
	if(bar.type == "spell" ) then
		self:ChangeSpell(bar);
	elseif(bar.type == "item") then
		self:ChangeItem(bar);
	elseif(bar.type == "macro") then
		self:ChangeMacro(bar);
	elseif(bar.type == nil) then
		self:ChangeEmpty(bar);
	end
	ClearCursor();
end


--Change TYPES
-- =========================
function TAB:ChangeSpell(bar)
	PickupSpell(bar.id);
	if( GetCursorInfo() ~= bar.type ) then
		self:Print("|cffff0000Unable to restore spell \"" .. bar.name .. "(" .. bar.rank .. ")\"  to slot no." .. bar.index ..", it does not appear to have been learned yet.");
		ClearCursor();
		return;
	end
	PlaceAction(bar.index);
end
function TAB:ChangeItem(bar)
	PickupItem(bar.id);
	if( GetCursorInfo() ~= bar.type ) then
		self:Print("|cffff0000Unable to restore item \"" .. bar.name .. "\" to slot no." .. bar.index .. ", cannot be found in inventory.");
		ClearCursor();
		return;
	end
	PlaceAction(bar.index);
end
function TAB:ChangeMacro(bar)
	if(not TAB.db.global.ignoreMacro and not TAB:CompareMacro(bar)) then
		local newId = TAB:FindMacro(bar);
		if(bar.id  == nil) then
			self:Print("|cffff0000Unable to restore macro with text \"" .. bar.macro .. "\" to slot no." .. bar.index .. ", it appears to have been deleted.");
			return;
		else
			bar.id = newId;
		end
	end
	PickupMacro(bar.id);
	if(GetCursorInfo() ~= bar.type) then
		self:Print("|cffff0000Unable to restore macro with text \"" .. bar.macro .. "\" to slot no." .. bar.index .. ", it appears to have been deleted.");
		ClearCursor();
		return;
	end
	PlaceAction(bar.index);
end
function TAB:ChangeEmpty(bar)
	PickupAction(bar.index);
	ClearCursor();
end


--MACRO Functions
-- =========================
function TAB:CompareMacro(node)
	local name, icon, macro = GetMacroInfo(node.id);
	if(node.macro == nil) then
		return true;
	end
	if(node.name == name and node.icon == icon and node.macro == macro) then
		return true;
	else
		return false;
	end
end
function TAB:FindMacro(node)
	for i = 1, 138 do
		local name, icon, macro = GetMacroInfo(i);
		if(node.name == name and node.icon == icon and node.macro == macro) then
			return i;
		end
	end
	return nil;
end


--TABLE FUNCTIONS
-- =========================
function TAB:FindIndex(tbl, itm)
	for i = 1, #tbl do
		if(tbl[i] == itm) then
			return i;
		end
	end
	return nil;
end
function TAB:IndexOfTalentString(itm)
	for i = 1, #TAB.db.char.talentStrings do
		if(TAB.db.char.talentStrings[i] == itm) then
			return i;
		end
	end
	return nil;
end

--Bar Naming
-- =========================
function TAB:FillNames()
	if(TAB.db.char.barNames == nil) then
		TAB.db.char.barNames = {};
	end
	if(#TAB.db.char.bars > #TAB.db.char.barNames) then
		for i = 1, #TAB.db.char.bars do
			if(#TAB.db.char.barNames == i - 1) then
				table.insert(TAB.db.char.barNames, "Bars #" .. i);
			elseif(TAB.db.char.barNames[i] == nil) then
				TAB.db.char.barNames[i] = "Bars #" .. i;
			end
		end
	end
end


--Version Check
-- =========================
function TAB:CheckVersion()
	if(GetAddOnMetadata("TalentActionBars", "Version") ~= nil) then
		if(self.db.global.version ~= GetAddOnMetadata("TalentActionBars", "Version")) then
			self:Print("Version of addon changed.");
			self.db.global.version = GetAddOnMetadata("TalentActionBars", "Version");
		end
	end
	TAB:SendBroadcastVersionCheck();
end


--Version Check Broadcast
-- =========================
function TAB:SendBroadcastVersionCheck()
	local serialized = self:Serialize(GetAddOnMetadata("TalentActionBars", "Version"));
	self:SendCommMessage("TABVersionCheck", serialized, "GUILD");
end
function TAB:OnCommReceived(prefix, message, distribution, sender)
	local success = false;
	local data;
	if(prefix == "TABVersionCheck") then
		success, data = self:Deserialize(message);
		if(success) then
			if(sender ~= UnitName("player")) then
				TAB:BroadcastVersionCheck(data);
			end
		end
	end
end
function TAB:BroadcastVersionCheck(version)
	if(tonumber(GetAddOnMetadata("TalentActionBars", "Version")) < tonumber(version) and self.versionWarrning == false) then
		self:Print("Newer version is availible!");
		self.versionWarrning = true;
	end
end

--Debug section
-- =========================
function TAB:Debug(text)
	if(self.db.global.debug) then
		TAB:Print("|cffffff00" .. tostring(text));
	end
end


function TAB:Test()
print("Test");
	
end